@@ -2,7 +2,7 @@
 /* $OpenLDAP$ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 2016-2021 The OpenLDAP Foundation.
+ * Copyright 2016-2026 The OpenLDAP Foundation.
  * Portions Copyright 2016 Symas Corporation.
  * All rights reserved.
  *
@@ -456,11 +456,21 @@ static ConfigOCs a_metaocs[] = {
 static int
 asyncmeta_ldadd( CfEntryInfo *p, Entry *e, ConfigArgs *c )
 {
+	a_metainfo_t	*mi;
+
 	if ( p->ce_type != Cft_Database || !p->ce_be ||
 		p->ce_be->be_cf_ocs != a_metaocs )
 		return LDAP_CONSTRAINT_VIOLATION;
 
 	c->be = p->ce_be;
+	mi = ( a_metainfo_t * )c->be->be_private;
+
+	if ( asyncmeta_db_has_pending_ops ( mi ) > 0 ) {
+		snprintf( c->cr_msg, sizeof( c->cr_msg ),
+			  "cannot modify a working database" );
+		Debug( LDAP_DEBUG_ANY, "%s: %s.\n", c->log, c->cr_msg );
+		return 1;
+	}
 	return LDAP_SUCCESS;
 }
 
@@ -486,15 +496,14 @@ asyncmeta_cfadd( Operation *op, SlapReply *rs, Entry *p, ConfigArgs *c )
 
 static int
 asyncmeta_back_new_target(
-	a_metatarget_t	**mtp )
+	a_metatarget_t	*mt,
+	a_metainfo_t     *mi,
+	BackendDB        *db )
 {
-	a_metatarget_t		*mt;
+	int i;
 
-	*mtp = NULL;
-
-	mt = ch_calloc( sizeof( a_metatarget_t ), 1 );
-
-	ldap_pvt_thread_mutex_init( &mt->mt_uri_mutex );
+	assert ( mi != NULL );
+	assert ( mt != NULL );
 
 	mt->mt_idassert_mode = LDAP_BACK_IDASSERT_LEGACY;
 	mt->mt_idassert_authmethod = LDAP_AUTH_NONE;
@@ -502,8 +511,21 @@ asyncmeta_back_new_target(
 	/* by default, use proxyAuthz control on each operation */
 	mt->mt_idassert_flags = LDAP_BACK_AUTH_PRESCRIPTIVE;
 
-	*mtp = mt;
-
+	if ( !SLAP_DBOPEN(db) || !(slapMode & SLAP_SERVER_MODE)) {
+		return 0;
+	}
+	for ( i = 0; i < mi->mi_num_conns; i++ ) {
+		a_metaconn_t *mc = &mi->mi_conns[i];
+		mc->mc_conns = ch_realloc( mc->mc_conns, sizeof( a_metasingleconn_t ) * mi->mi_ntargets);
+		memset( &(mc->mc_conns[mi->mi_ntargets-1]), 0, sizeof( a_metasingleconn_t ) );
+	}
+	/* If this is the first target, start the timeout loop */
+	if ( mi->mi_ntargets == 1 ) {
+		ldap_pvt_thread_mutex_lock( &slapd_rq.rq_mutex );
+		mi->mi_task = ldap_pvt_runqueue_insert( &slapd_rq, 1,
+							asyncmeta_timeout_loop, mi, "asyncmeta_timeout_loop", mi->mi_suffix.bv_val );
+		ldap_pvt_thread_mutex_unlock( &slapd_rq.rq_mutex );
+	}
 	return 0;
 }
 
@@ -1026,6 +1048,13 @@ asyncmeta_back_cf_gen( ConfigArgs *c )
 		}
 	}
 
+	if ( c->op != SLAP_CONFIG_EMIT && asyncmeta_db_has_pending_ops ( mi ) > 0 ) {
+		snprintf( c->cr_msg, sizeof( c->cr_msg ),
+			  "cannot modify a working database" );
+		Debug( LDAP_DEBUG_ANY, "%s: %s.\n", c->log, c->cr_msg );
+		return 1;
+	}
+
 	if ( c->op == SLAP_CONFIG_EMIT ) {
 		struct berval bv = BER_BVNULL;
 
@@ -1520,6 +1549,7 @@ asyncmeta_back_cf_gen( ConfigArgs *c )
 		}
 		return rc;
 	} else if ( c->op == LDAP_MOD_DELETE ) {
+
 		switch( c->type ) {
 		/* Base attrs */
 		case LDAP_BACK_CFG_DNCACHE_TTL:
@@ -1544,8 +1574,15 @@ asyncmeta_back_cf_gen( ConfigArgs *c )
 
 		/* common attrs */
 		case LDAP_BACK_CFG_BIND_TIMEOUT:
-			mc->mc_bind_timeout.tv_sec = 0;
-			mc->mc_bind_timeout.tv_usec = 0;
+			if ( asyncmeta_db_has_mscs ( mi ) > 0 ) {
+				snprintf( c->cr_msg, sizeof( c->cr_msg ),
+					  "cannot modify this attribute if there are established target connections" );
+				Debug( LDAP_DEBUG_ANY, "%s: %s.\n", c->log, c->cr_msg );
+				rc = 1;
+			} else {
+				mc->mc_bind_timeout.tv_sec = 0;
+				mc->mc_bind_timeout.tv_usec = 0;
+			}
 			break;
 
 		case LDAP_BACK_CFG_CANCEL:
@@ -1567,7 +1604,14 @@ asyncmeta_back_cf_gen( ConfigArgs *c )
 			break;
 
 		case LDAP_BACK_CFG_NETWORK_TIMEOUT:
-			mc->mc_network_timeout = 0;
+			if ( asyncmeta_db_has_mscs ( mi ) > 0 ) {
+				snprintf( c->cr_msg, sizeof( c->cr_msg ),
+					  "cannot modify this attribute if there are established target connections" );
+				Debug( LDAP_DEBUG_ANY, "%s: %s.\n", c->log, c->cr_msg );
+				rc = 1;
+			} else {
+				mc->mc_network_timeout = 0;
+			}
 			break;
 
 		case LDAP_BACK_CFG_NOREFS:
@@ -1604,7 +1648,14 @@ asyncmeta_back_cf_gen( ConfigArgs *c )
 			break;
 
 		case LDAP_BACK_CFG_VERSION:
-			mc->mc_version = 0;
+			if ( asyncmeta_db_has_mscs ( mi ) > 0 ) {
+				snprintf( c->cr_msg, sizeof( c->cr_msg ),
+					  "cannot modify this attribute if there are established target connections" );
+				Debug( LDAP_DEBUG_ANY, "%s: %s.\n", c->log, c->cr_msg );
+				rc = 1;
+			} else {
+				mc->mc_version = 0;
+			}
 			break;
 
 #ifdef SLAP_CONTROL_X_SESSION_TRACKING
@@ -1625,9 +1676,16 @@ asyncmeta_back_cf_gen( ConfigArgs *c )
 
 		/* target attrs */
 		case LDAP_BACK_CFG_URI:
-			if ( mt->mt_uri ) {
-				ch_free( mt->mt_uri );
-				mt->mt_uri = NULL;
+			if ( asyncmeta_db_has_mscs ( mi ) > 0 ) {
+				snprintf( c->cr_msg, sizeof( c->cr_msg ),
+					  "cannot modify this attribute if there are established target connections" );
+				Debug( LDAP_DEBUG_ANY, "%s: %s.\n", c->log, c->cr_msg );
+				rc = 1;
+			} else {
+				if ( mt->mt_uri ) {
+					ch_free( mt->mt_uri );
+					mt->mt_uri = NULL;
+				}
 			}
 			/* FIXME: should have a way to close all cached
 			 * connections associated with this target.
@@ -1638,44 +1696,66 @@ asyncmeta_back_cf_gen( ConfigArgs *c )
 			BerVarray *bvp;
 
 			bvp = &mt->mt_idassert_authz;
-			if ( c->valx < 0 ) {
-				if ( *bvp != NULL ) {
-					ber_bvarray_free( *bvp );
-					*bvp = NULL;
-				}
-
+			if ( asyncmeta_db_has_mscs ( mi ) > 0 ) {
+				snprintf( c->cr_msg, sizeof( c->cr_msg ),
+					  "cannot modify this attribute if there are established target connections" );
+				Debug( LDAP_DEBUG_ANY, "%s: %s.\n", c->log, c->cr_msg );
+				rc = 1;
 			} else {
-				if ( *bvp == NULL ) {
-					rc = 1;
-					break;
-				}
+				if ( c->valx < 0 ) {
+					if ( *bvp != NULL ) {
+						ber_bvarray_free( *bvp );
+						*bvp = NULL;
+					}
 
-				for ( i = 0; !BER_BVISNULL( &((*bvp)[ i ]) ); i++ )
-					;
+				} else {
+					if ( *bvp == NULL ) {
+						rc = 1;
+						break;
+					}
 
-				if ( i >= c->valx ) {
-					rc = 1;
-					break;
+					for ( i = 0; !BER_BVISNULL( &((*bvp)[ i ]) ); i++ )
+						;
+
+					if ( i >= c->valx ) {
+						rc = 1;
+						break;
+					}
+					ber_memfree( ((*bvp)[ c->valx ]).bv_val );
+					for ( i = c->valx; !BER_BVISNULL( &((*bvp)[ i + 1 ]) ); i++ ) {
+						(*bvp)[ i ] = (*bvp)[ i + 1 ];
+					}
+					BER_BVZERO( &((*bvp)[ i ]) );
 				}
-				ber_memfree( ((*bvp)[ c->valx ]).bv_val );
-				for ( i = c->valx; !BER_BVISNULL( &((*bvp)[ i + 1 ]) ); i++ ) {
-					(*bvp)[ i ] = (*bvp)[ i + 1 ];
-				}
-				BER_BVZERO( &((*bvp)[ i ]) );
 			}
-			} break;
+		}
+			break;
 
 		case LDAP_BACK_CFG_IDASSERT_BIND:
-			bindconf_free( &mt->mt_idassert.si_bc );
-			memset( &mt->mt_idassert, 0, sizeof( slap_idassert_t ) );
+			if ( asyncmeta_db_has_mscs ( mi ) > 0 ) {
+				snprintf( c->cr_msg, sizeof( c->cr_msg ),
+					  "cannot modify this attribute if there are established target connections" );
+				Debug( LDAP_DEBUG_ANY, "%s: %s.\n", c->log, c->cr_msg );
+				rc = 1;
+			} else {
+				bindconf_free( &mt->mt_idassert.si_bc );
+				memset( &mt->mt_idassert, 0, sizeof( slap_idassert_t ) );
+			}
 			break;
 
 		case LDAP_BACK_CFG_SUFFIXM:
-			if ( mt->mt_lsuffixm.bv_val ) {
-				ch_free( mt->mt_lsuffixm.bv_val );
-				ch_free( mt->mt_rsuffixm.bv_val );
-				BER_BVZERO( &mt->mt_lsuffixm );
-				BER_BVZERO( &mt->mt_rsuffixm );
+			if ( asyncmeta_db_has_mscs ( mi ) > 0 ) {
+				snprintf( c->cr_msg, sizeof( c->cr_msg ),
+					  "cannot modify this attribute if there are established target connections" );
+				Debug( LDAP_DEBUG_ANY, "%s: %s.\n", c->log, c->cr_msg );
+				rc = 1;
+			} else {
+				if ( mt->mt_lsuffixm.bv_val ) {
+					ch_free( mt->mt_lsuffixm.bv_val );
+					ch_free( mt->mt_rsuffixm.bv_val );
+					BER_BVZERO( &mt->mt_lsuffixm );
+					BER_BVZERO( &mt->mt_rsuffixm );
+				}
 			}
 			break;
 
@@ -1729,7 +1809,10 @@ asyncmeta_back_cf_gen( ConfigArgs *c )
 			break;
 
 		case LDAP_BACK_CFG_MAX_TARGET_CONNS:
-			mi->mi_max_target_conns = 0;
+			snprintf( c->cr_msg, sizeof( c->cr_msg ),
+				  "max-target-conns cannot be modified at runtime" );
+			Debug( LDAP_DEBUG_ANY, "%s: %s.\n", c->log, c->cr_msg );
+			rc = 1;
 			break;
 
 		case LDAP_BACK_CFG_MAX_TIMEOUT_OPS:
@@ -1737,9 +1820,16 @@ asyncmeta_back_cf_gen( ConfigArgs *c )
 			break;
 
 		case LDAP_BACK_CFG_KEEPALIVE:
-			mt->mt_tls.sb_keepalive.sk_idle = 0;
-			mt->mt_tls.sb_keepalive.sk_probes = 0;
-			mt->mt_tls.sb_keepalive.sk_interval = 0;
+			if ( asyncmeta_db_has_mscs ( mi ) > 0 ) {
+				snprintf( c->cr_msg, sizeof( c->cr_msg ),
+					  "cannot modify this attribute if there are established target connections" );
+				Debug( LDAP_DEBUG_ANY, "%s: %s.\n", c->log, c->cr_msg );
+				rc = 1;
+			} else {
+				mt->mt_tls.sb_keepalive.sk_idle = 0;
+				mt->mt_tls.sb_keepalive.sk_probes = 0;
+				mt->mt_tls.sb_keepalive.sk_interval = 0;
+			}
 			break;
 
 		case LDAP_BACK_CFG_TCP_USER_TIMEOUT:
@@ -1791,7 +1881,6 @@ asyncmeta_back_cf_gen( ConfigArgs *c )
 		LDAPURLDesc 	*ludp;
 		struct berval	dn;
 		int		j;
-
 		char		**uris = NULL;
 
 		if ( c->be->be_nsuffix == NULL ) {
@@ -1801,39 +1890,6 @@ asyncmeta_back_cf_gen( ConfigArgs *c )
 			return 1;
 		}
 
-		i = mi->mi_ntargets++;
-
-		mi->mi_targets = ( a_metatarget_t ** )ch_realloc( mi->mi_targets,
-			sizeof( a_metatarget_t * ) * mi->mi_ntargets );
-		if ( mi->mi_targets == NULL ) {
-			snprintf( c->cr_msg, sizeof( c->cr_msg ),
-				"out of memory while storing server name"
-				" in \"%s <protocol>://<server>[:port]/<naming context>\"",
-				c->argv[0] );
-			Debug( LDAP_DEBUG_ANY, "%s: %s.\n", c->log, c->cr_msg );
-			return 1;
-		}
-
-		if ( asyncmeta_back_new_target( &mi->mi_targets[ i ] ) != 0 ) {
-			snprintf( c->cr_msg, sizeof( c->cr_msg ),
-				"unable to init server"
-				" in \"%s <protocol>://<server>[:port]/<naming context>\"",
-				c->argv[0] );
-			Debug( LDAP_DEBUG_ANY, "%s: %s.\n", c->log, c->cr_msg );
-			return 1;
-		}
-
-		mt = mi->mi_targets[ i ];
-
-		mt->mt_rebind_f = mi->mi_rebind_f;
-		mt->mt_urllist_f = mi->mi_urllist_f;
-		mt->mt_urllist_p = mt;
-
-		if ( META_BACK_QUARANTINE( mi ) ) {
-			ldap_pvt_thread_mutex_init( &mt->mt_quarantine_mutex );
-		}
-		mt->mt_mc = mi->mi_mc;
-
 		for ( j = 1; j < c->argc; j++ ) {
 			char	**tmpuris = ldap_str2charray( c->argv[ j ], "\t" );
 
@@ -1855,6 +1911,8 @@ asyncmeta_back_cf_gen( ConfigArgs *c )
 			}
 		}
 
+		mt = ch_calloc( sizeof( a_metatarget_t ), 1 );
+
 		for ( j = 0; uris[ j ] != NULL; j++ ) {
 			char *tmpuri = NULL;
 
@@ -1871,6 +1929,7 @@ asyncmeta_back_cf_gen( ConfigArgs *c )
 					j-1, c->argv[0] );
 				Debug( LDAP_DEBUG_ANY, "%s: %s.\n", c->log, c->cr_msg );
 				ldap_charray_free( uris );
+				asyncmeta_target_free( mt );
 				return 1;
 			}
 
@@ -1887,6 +1946,7 @@ asyncmeta_back_cf_gen( ConfigArgs *c )
 					Debug( LDAP_DEBUG_ANY, "%s: %s.\n", c->log, c->cr_msg );
 					ldap_free_urllist( ludp );
 					ldap_charray_free( uris );
+					asyncmeta_target_free( mt );
 					return 1;
 				}
 
@@ -1903,6 +1963,7 @@ asyncmeta_back_cf_gen( ConfigArgs *c )
 					Debug( LDAP_DEBUG_ANY, "%s: %s.\n", c->log, c->cr_msg );
 					ldap_free_urllist( ludp );
 					ldap_charray_free( uris );
+					asyncmeta_target_free( mt );
 					return( 1 );
 				}
 
@@ -1925,6 +1986,7 @@ asyncmeta_back_cf_gen( ConfigArgs *c )
 					Debug( LDAP_DEBUG_ANY, "%s: %s.\n", c->log, c->cr_msg );
 					ldap_free_urllist( ludp );
 					ldap_charray_free( uris );
+					asyncmeta_target_free( mt );
 					return( 1 );
 				}
 
@@ -1936,6 +1998,7 @@ asyncmeta_back_cf_gen( ConfigArgs *c )
 					Debug( LDAP_DEBUG_ANY, "%s: %s.\n", c->log, c->cr_msg );
 					ldap_free_urllist( ludp );
 					ldap_charray_free( uris );
+					asyncmeta_target_free( mt );
 					return( 1 );
 
 				}
@@ -1947,6 +2010,7 @@ asyncmeta_back_cf_gen( ConfigArgs *c )
 				snprintf( c->cr_msg, sizeof( c->cr_msg ), "no memory?" );
 				Debug( LDAP_DEBUG_ANY, "%s: %s.\n", c->log, c->cr_msg );
 				ldap_charray_free( uris );
+				asyncmeta_target_free( mt );
 				return( 1 );
 			}
 			ldap_memfree( uris[ j ] );
@@ -1958,7 +2022,10 @@ asyncmeta_back_cf_gen( ConfigArgs *c )
 		if ( mt->mt_uri == NULL) {
 			snprintf( c->cr_msg, sizeof( c->cr_msg ), "no memory?" );
 			Debug( LDAP_DEBUG_ANY, "%s: %s.\n", c->log, c->cr_msg );
+			asyncmeta_target_free( mt );
 			return( 1 );
+		} else {
+			ldap_pvt_thread_mutex_init( &mt->mt_uri_mutex );
 		}
 
 		/*
@@ -1974,8 +2041,47 @@ asyncmeta_back_cf_gen( ConfigArgs *c )
 			snprintf( c->cr_msg, sizeof( c->cr_msg ),
 				"<naming context> of URI must be within the naming context of this database." );
 			Debug( LDAP_DEBUG_ANY, "%s: %s.\n", c->log, c->cr_msg );
+			asyncmeta_target_free( mt );
 			return 1;
 		}
+
+		i = mi->mi_ntargets++;
+
+		mi->mi_targets = ( a_metatarget_t ** )ch_realloc( mi->mi_targets,
+			sizeof( a_metatarget_t * ) * mi->mi_ntargets );
+		if ( mi->mi_targets == NULL ) {
+			snprintf( c->cr_msg, sizeof( c->cr_msg ),
+				"out of memory while storing server name"
+				" in \"%s <protocol>://<server>[:port]/<naming context>\"",
+				c->argv[0] );
+			Debug( LDAP_DEBUG_ANY, "%s: %s.\n", c->log, c->cr_msg );
+			asyncmeta_target_free( mt );
+			return 1;
+		}
+
+		mi->mi_targets[ i ] = mt;
+		if ( asyncmeta_back_new_target( mt, mi, c->be ) != 0 ) {
+			/* does not currently happen, asyncmeta back_new_target always returns 0 */
+			snprintf( c->cr_msg, sizeof( c->cr_msg ),
+				"unable to init server"
+				" in \"%s <protocol>://<server>[:port]/<naming context>\"",
+				c->argv[0] );
+			Debug( LDAP_DEBUG_ANY, "%s: %s.\n", c->log, c->cr_msg );
+			asyncmeta_target_free( mt );
+			mi->mi_ntargets--;
+			return 1;
+		}
+
+
+		mt->mt_rebind_f = mi->mi_rebind_f;
+		mt->mt_urllist_f = mi->mi_urllist_f;
+		mt->mt_urllist_p = mt;
+
+		if ( META_BACK_QUARANTINE( mi ) ) {
+			ldap_pvt_thread_mutex_init( &mt->mt_quarantine_mutex );
+		}
+		mt->mt_mc = mi->mi_mc;
+
 		c->ca_private = mt;
 		config_push_cleanup( c, asyncmeta_cf_cleanup );
 	} break;
@@ -2405,9 +2511,11 @@ asyncmeta_back_cf_gen( ConfigArgs *c )
 		break;
 #endif /* SLAPD_META_CLIENT_PR */
 
-	case LDAP_BACK_CFG_KEEPALIVE:
-		slap_keepalive_parse( ber_bvstrdup(c->argv[1]),
-				 &mt->mt_tls.sb_keepalive, 0, 0, 0);
+	case LDAP_BACK_CFG_KEEPALIVE: {
+		struct berval bv;
+		ber_str2bv( c->argv[1], 0, 1, &bv );
+		slap_keepalive_parse( &bv, &mt->mt_tls.sb_keepalive, 0, 0, 0 );
+		}
 		break;
 
 	case LDAP_BACK_CFG_TCP_USER_TIMEOUT:
