@@ -136,6 +136,7 @@ static AttributeDescription	*ad_memberOf;
 static ObjectClass			*oc_group;
 
 static slap_overinst		memberof;
+static slap_overinst		*oi_refint;
 
 typedef struct memberof_t {
 	struct berval		mo_dn;
@@ -574,6 +575,7 @@ memberof_addcheck( Operation *op )
 	o.o_req_dn = op->o_bd->be_suffix[0];
 	o.o_req_ndn = op->o_bd->be_nsuffix[0];
 	o.o_do_not_cache = 1;
+	o.o_managedsait = SLAP_CONTROL_CRITICAL;
 	o.ors_scope = LDAP_SCOPE_SUBTREE;
 	o.ors_slimit = SLAP_NO_LIMIT;
 	o.ors_tlimit = SLAP_NO_LIMIT;
@@ -905,6 +907,8 @@ memberof_op_modify( Operation *op, SlapReply *rs )
 	LDAP_SLIST_FOREACH( oex, &op->o_extra, oe_next ) {
 		if ( oex->oe_key == (void *)&memberof )
 			return SLAP_CB_CONTINUE;
+		if ( oi_refint && oex->oe_key == oi_refint )
+			return SLAP_CB_CONTINUE;
 	}
 
 	if ( MEMBEROF_REVERSE( mo ) ) {
@@ -1300,10 +1304,8 @@ done2:;
 	if ( save_member ) {
 		op->o_dn = op->o_bd->be_rootdn;
 		op->o_ndn = op->o_bd->be_rootndn;
-		op->o_bd->bd_info = (BackendInfo *)on->on_info;
 		rc = backend_attribute( op, NULL, &op->o_req_ndn,
 				mo->mo_ad_member, &mci->member, ACL_READ );
-		op->o_bd->bd_info = (BackendInfo *)on;
 	}
 
 	sc->sc_next = op->o_callback;
@@ -1502,10 +1504,8 @@ memberof_res_modify( Operation *op, SlapReply *rs )
 
 		case LDAP_MOD_REPLACE:
 			/* delete all ... */
-			op->o_bd->bd_info = (BackendInfo *)on->on_info;
 			rc = backend_attribute( op, NULL, &op->o_req_ndn,
 					mo->mo_ad_memberof, &vals, ACL_READ );
-			op->o_bd->bd_info = (BackendInfo *)on;
 			if ( rc == LDAP_SUCCESS ) {
 				for ( i = 0; !BER_BVISNULL( &vals[ i ] ); i++ ) {
 					memberof_value_modify( op,
@@ -1640,10 +1640,8 @@ memberof_res_modrdn( Operation *op, SlapReply *rs )
 	}
 
 	if ( mci->what & MEMBEROF_IS_GROUP ) {
-		op->o_bd->bd_info = (BackendInfo *)on->on_info;
 		rc = backend_attribute( op, NULL, &op->orr_nnewDN,
 				mo->mo_ad_member, &vals, ACL_READ );
-		op->o_bd->bd_info = (BackendInfo *)on;
 
 		if ( rc == LDAP_SUCCESS ) {
 			for ( i = 0; !BER_BVISNULL( &vals[ i ] ); i++ ) {
@@ -1657,10 +1655,8 @@ memberof_res_modrdn( Operation *op, SlapReply *rs )
 	}
 
 	if ( MEMBEROF_REFINT( mo ) && ( mci->what & MEMBEROF_IS_MEMBER ) ) {
-		op->o_bd->bd_info = (BackendInfo *)on->on_info;
 		rc = backend_attribute( op, NULL, &op->orr_nnewDN,
 				mo->mo_ad_memberof, &vals, ACL_READ );
-		op->o_bd->bd_info = (BackendInfo *)on;
 
 		if ( rc == LDAP_SUCCESS ) {
 			for ( i = 0; !BER_BVISNULL( &vals[ i ] ); i++ ) {
@@ -1688,11 +1684,6 @@ memberof_db_init(
 	const char		*text = NULL;
 	int rc;
 
-	mo = (memberof_t *)ch_calloc( 1, sizeof( memberof_t ) );
-
-	/* safe default */
-	mo->mo_dangling_err = LDAP_CONSTRAINT_VIOLATION;
-
 	if ( !ad_memberOf ) {
 		rc = slap_str2ad( SLAPD_MEMBEROF_ATTR, &ad_memberOf, &text );
 		if ( rc != LDAP_SUCCESS ) {
@@ -1724,6 +1715,11 @@ memberof_db_init(
 		}
 	}
 
+	mo = (memberof_t *)ch_calloc( 1, sizeof( memberof_t ) );
+
+	/* safe default */
+	mo->mo_dangling_err = LDAP_CONSTRAINT_VIOLATION;
+
 	on->on_bi.bi_private = (void *)mo;
 
 	return 0;
@@ -2158,6 +2154,14 @@ mo_cf_gen( ConfigArgs *c )
 
 		case MO_ADDCHECK:
 			if ( c->value_int ) {
+				if ( SLAP_ISGLOBALOVERLAY( c->be ) ) {
+					snprintf( c->cr_msg, sizeof( c->cr_msg ),
+						"addcheck functionality not supported "
+						"when memberof is a global overlay" );
+					Debug( LDAP_DEBUG_ANY, "%s: %s.\n",
+						c->log, c->cr_msg );
+					return 1;
+				}
 				mo->mo_flags |= MEMBEROF_FADDCHECK;
 
 			} else {
@@ -2209,6 +2213,8 @@ memberof_db_open(
 		memberof_make_member_filter( mo );
 	}
 
+	oi_refint = overlay_find( "refint" );
+
 	return 0;
 }
 
